@@ -15,7 +15,9 @@ template <class Key, class T> class hash_map : public std::unordered_map<Key, T>
 
 typedef double value_t_ripser;
 typedef int64_t index_t_ripser;
-typedef int16_t coefficient_t_ripser;
+typedef uint8_t coefficient_t_ripser;
+
+static const size_t num_coefficient_bits = 8;
 
 class binomial_coeff_table {
   std::vector<std::vector<index_t_ripser>> B;
@@ -97,11 +99,22 @@ std::vector<index_t_ripser> vertices_of_simplex(const index_t_ripser simplex_ind
   return vertices;
 }
 
-typedef index_t_ripser entry_t;
-const index_t_ripser get_index(entry_t i) { return i; }
-index_t_ripser get_coefficient(entry_t i) { return 1; }
-entry_t make_entry(index_t_ripser _index, coefficient_t_ripser _value) { return entry_t(_index); }
-void set_coefficient(index_t_ripser& e, const coefficient_t_ripser c) {}
+#pragma pack(1)
+struct entry_t {
+ index_t_ripser index : 8 * sizeof(index_t_ripser) - num_coefficient_bits;
+ coefficient_t_ripser coefficient : num_coefficient_bits;
+ entry_t(index_t_ripser _index, coefficient_t_ripser _coefficient) : index(_index), coefficient(_coefficient) {}
+ entry_t(index_t_ripser _index) : index(_index), coefficient(0) {}
+ entry_t() : index(0), coefficient(0) {}
+};
+#pragma pack() // reset 
+
+//  static_assert(sizeof(entry_t) == sizeof(index_t_ripser), "size of entry_t is not the same as index_t");
+
+entry_t make_entry(index_t_ripser i, coefficient_t_ripser c) { return entry_t(i, c); }
+index_t_ripser get_index(const entry_t& e) { return e.index; }
+index_t_ripser get_coefficient(const entry_t& e) { return e.coefficient; }
+void set_coefficient(entry_t& e, const coefficient_t_ripser c) { e.coefficient = c; }
 
 const entry_t& get_entry(const entry_t& e) { return e; }
 
@@ -517,7 +530,7 @@ template <typename T> T read(std::istream& s) {
   return result; // on little endian: boost::endian::little_to_native(result);
 }
 
-compressed_lower_distance_matrix getPointCloud(NumericMatrix inputMat) {
+compressed_lower_distance_matrix getPointCloud(const NumericMatrix& inputMat) {
   std::vector<std::vector<value_t_ripser>> points;
 
   int numRows = inputMat.nrow(),
@@ -549,7 +562,7 @@ compressed_lower_distance_matrix getPointCloud(NumericMatrix inputMat) {
   return compressed_lower_distance_matrix(std::move(distances));
 }
 
-compressed_lower_distance_matrix getLowerDistMatrix(NumericMatrix inputMat) {
+compressed_lower_distance_matrix getLowerDistMatrix(const NumericMatrix& inputMat) {
 	std::vector<value_t_ripser> distances;
 	value_t_ripser value;
 
@@ -569,7 +582,7 @@ compressed_lower_distance_matrix getLowerDistMatrix(NumericMatrix inputMat) {
 }
 
 // convert from user format into lower distance matrix
-compressed_lower_distance_matrix read_file(NumericMatrix input_points, int format) {
+compressed_lower_distance_matrix read_file(const NumericMatrix& input_points, int format) {
 	switch (format) {
 		case 0:
 			return getPointCloud(input_points);
@@ -582,15 +595,11 @@ compressed_lower_distance_matrix read_file(NumericMatrix input_points, int forma
 	}
 }
 
-// Altered version of Ripser by Ulrich Bauer
-// format = 0 --> point cloud
-// format = 1 --> lower distance matrix
-// [[Rcpp::export]]
-NumericVector ripser_cpp(NumericMatrix input_points, int dim, float thresh, int format) {
-  NumericVector ansx(1);
-  ansx[0] = 1;
-
-  //MY VARS
+// Given distances and parameters, computes barcodes 
+template < typename DistanceMatrix >
+NumericVector ripser_compute(const DistanceMatrix& dist, int dim, float thresh, int p){
+ 
+   //MY VARS
   int currDim = 0;
   std::vector<std::vector<value_t_ripser>> pers_hom;
 
@@ -598,14 +607,11 @@ NumericVector ripser_cpp(NumericMatrix input_points, int dim, float thresh, int
   value_t_ripser threshold = std::numeric_limits<value_t_ripser>::max();
   if (thresh > 0)
     threshold = thresh;
-  const coefficient_t_ripser modulus = 2;
   
-  //make sure a valid format is used
-  assert(format == 0 || format == 1);
-  
-  //get distance matrix based on input format
-  compressed_lower_distance_matrix dist = read_file(input_points, format);
-
+  // MJP - Check coefficient p is prime (and positive). 
+  if (p < 0 || !is_prime(p)){ Rcpp::stop("Non-prime supplied to p."); }
+  const coefficient_t_ripser modulus = p;
+ 
   index_t_ripser n = dist.size();
   dim_max = std::min(dim_max, n - 2);
   binomial_coeff_table binomial_coeff(n, dim_max + 2);
@@ -668,11 +674,35 @@ NumericVector ripser_cpp(NumericMatrix input_points, int dim, float thresh, int
 
   NumericVector ans(pers_hom.size() * 3);
   int ind = 0;
-  for (int i = 0; i < pers_hom.size(); i++)
-    for (int j = 0; j < 3; j++)
-    {
+  for (int i = 0; i < pers_hom.size(); i++){
+   for (int j = 0; j < 3; j++) {
       ans[ind++] = pers_hom[i][j];
     }
+  } 
+  return(ans);
+}
+  
+  
+// [[Rcpp::export]]
+NumericVector ripser_cpp_dist(const NumericVector& dist_r, int dim, float thresh, int p){
+ std::vector< value_t_ripser > distances( dist_r.size() );
+ std::move(dist_r.begin(), dist_r.end(), distances.begin());
+ compressed_upper_distance_matrix dist = compressed_upper_distance_matrix(std::move(distances));
+ return ripser_compute(dist, dim, thresh, p);
+}
 
-    return ans;
+// Altered version of Ripser by Ulrich Bauer
+// format = 0 --> point cloud
+// format = 1 --> lower distance matrix
+// [[Rcpp::export]]
+NumericVector ripser_cpp(const NumericMatrix& input_points, int dim, float thresh, int p, int format) {
+
+  //make sure a valid format is used
+  assert(format == 0 || format == 1);
+  
+  //get distance matrix based on input format
+  compressed_lower_distance_matrix dist = read_file(input_points, format);
+  
+  // Return barcodes 
+  return ripser_compute(dist, dim, thresh, p);
 }
